package be.iminds.ilabt.jfed.highlevel.jobs;

import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.ExperimentPart;
import be.iminds.ilabt.jfed.experiment.SfaExperimentPart;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.highlevel.DesiredStatus;
import be.iminds.ilabt.jfed.highlevel.controller.TaskThread;
import be.iminds.ilabt.jfed.highlevel.jobs.parts.ExperimentPartController;
import be.iminds.ilabt.jfed.highlevel.jobs.parts.ExperimentPartControllerManager;
import be.iminds.ilabt.jfed.highlevel.jobs.parts.NonSfaExperimentPartController;
import be.iminds.ilabt.jfed.highlevel.jobs.sfa_parts.UpdateSfaExperimentPartUntilTargetOpStatusSlice;
import be.iminds.ilabt.jfed.highlevel.jobs.states.JobStateFactory;
import be.iminds.ilabt.jfed.highlevel.model.InternalState;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.highlevel.util.AggregateManagerWrapperFactory;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.RspecNode;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import java.util.*;
import java.util.stream.Collectors;

/**
 * Wait for a specific "geni_operational_status" in a "Status" reply. (Geni AM v3 case)
 * This is typically used to wait for a "geni_ready" status.
 * However, between Provision and POA start, it can also be needed to wait for a "geni_notready" status.
 *
 * This class is more AM generic, and can thus also wait for an AMv2 "geni_status": "ready"
 *
 * See also:
 * http://groups.geni.net/geni/wiki/GAPI_AM_API_V3/CommonConcepts#SliverOperationalStates
 * http://groups.geni.net/geni/wiki/GAPI_AM_API_V3/CommonConcepts#SliverOperationalActions
 */
public class WaitForOpStatusExperimentJob extends AbstractJob<Boolean> {
    private static final Logger LOG = LoggerFactory.getLogger(WaitForOpStatusExperimentJob.class);

    @Nonnull private final JobStateFactory jobStateFactory;

    @Nonnull private final AggregateManagerWrapperFactory aggregateManagerWrapperFactory;
    @Nonnull private final ExperimentPartControllerManager epcManager;

    private Boolean success = null;

    @Nonnull private final DesiredStatus desiredStatus;

    public WaitForOpStatusExperimentJob(@Nonnull Experiment experiment,
                                        @Nonnull HighLevelTaskFactory hltf,
                                        @Nonnull TaskThread tt,
                                        @Nonnull JobStateFactory jobStateFactory,
                                        @Nonnull AggregateManagerWrapperFactory aggregateManagerWrapperFactory,
                                        @Nonnull ExperimentPartControllerManager epcManager,
                                        @Nonnull DesiredStatus desiredStatus) {
        super("Wait for "+desiredStatus+ " for Experiment " + experiment.getName(),
                experiment, hltf, tt);

        this.jobStateFactory = jobStateFactory;
        this.aggregateManagerWrapperFactory = aggregateManagerWrapperFactory;
        this.epcManager = epcManager;

        this.desiredStatus = desiredStatus;

        assert experiment.getRequestedEndTime() != null;
    }

    @Override
    public Boolean execute() throws Exception {
        if (!hasSfaParts()) {
            if (experiment.getExperimentSpecification() != null) {
                experiment.getExperimentSpecificationLogger().firePostRSpecFail(experiment.getExperimentSpecification().getRspecs().get(0),
                        "assert failed: experiment.getExperimentSpecification() == null", null);
                experiment.getExperimentSpecificationLogger().firePostRspecAll(false, Collections.emptyList());
            }
            return true;
        }

        boolean permanentFailure = false;

        LOG.debug("WaitForOpStatusExperimentJob started");

        final UpdateSfaPartsUntilTargetOpStatus updateUntilReadyState =
                new UpdateSfaPartsUntilTargetOpStatus(jobStateFactory, this, desiredStatus);
        setAndRunState(updateUntilReadyState);

        if (updateUntilReadyState.isPermanentFailure()) {
            permanentFailure = true;
            LOG.info("WaitForPartsReadyState reported permanent failure");
        }

        if (permanentFailure) {
            if (experiment.getExperimentSpecification() != null) {
                experiment.getExperimentSpecificationLogger().firePostRSpecFail(
                        experiment.getExperimentSpecification().getRspecs().get(0),
                        "WaitForPartsReadyState reported permanent failure", null);
                experiment.getExperimentSpecificationLogger().firePostRspecAll(
                        false, getExperimentClientIds());
            }

            LOG.debug("WaitForOpStatusExperimentJob stopped in failure");
            success = false;
            return false;
        }

        LOG.debug("WaitForOpStatusExperimentJob stopped successfully");

        success = true;
        return true;
    }

    @Nonnull
    private List<String> getExperimentClientIds() {
        if (experiment.getSliceOrNull() == null || experiment.getSliceOrNull().getRequestRspec() == null) {
            return Collections.emptyList();
        }
        RequestRspecSource requestRspecSource = experiment.getSliceOrNull().getRequestRspec();
        FXModelRspec model = requestRspecSource == null ?
                null :
                requestRspecSource.getModelRspec(ModelRspecType.FX);
        assert model != null;

        if (model == null) {
            return Collections.emptyList();
        }

        return model.getNodes().stream().map(RspecNode::getClientId).filter(Objects::nonNull).collect(Collectors.toList());
    }

    public Boolean isSuccess() {
        return success;
    }

    public static class UpdateSfaPartsUntilTargetOpStatus extends SlicedState<UpdateSfaExperimentPartUntilTargetOpStatusSlice> {
        private final List<UpdateSfaExperimentPartUntilTargetOpStatusSlice> slices;
        private final Experiment experiment;
        private final JobStateFactory jobStateFactory;
        private final Job job;

        private final DesiredStatus desiredStatus;

        UpdateSfaPartsUntilTargetOpStatus(@Nonnull JobStateFactory jobStateFactory,
                                          @Nonnull Job job,
                                          @Nonnull DesiredStatus desiredStatus) {
            super("Waiting for resources to become ready");

            this.jobStateFactory = jobStateFactory;
            this.job = job;
            this.desiredStatus = desiredStatus;
            this.experiment = job.getExperiment();

            slices = experiment.getPartsStream()
                    .filter(part -> part.getState() != InternalState.FAILED)
                    .filter(part -> part instanceof SfaExperimentPart)
                    .map(part -> new UpdateSfaExperimentPartUntilTargetOpStatusSlice(
                            job,
                            (SfaExperimentPart) part,
                            experiment.getCheckReadyInterval(),
                            experiment.getMaxWaitUntilReady(),
                            jobStateFactory,
                            desiredStatus))
                    .collect(Collectors.toList());
        }


        public boolean isPermanentFailure() {
            return getSlices().stream().anyMatch(updateSlice -> updateSlice.getResult() != ExperimentTaskStatus.SUCCESS);
        }

        @Override
        public Collection<UpdateSfaExperimentPartUntilTargetOpStatusSlice> getSlices() {
            return slices;
        }
    }






    private Collection<SfaExperimentPart> getSfaParts() {
        return experiment.getPartsStream()
                .filter(part -> part instanceof SfaExperimentPart)
                .map(SfaExperimentPart.class::cast)
                .collect(Collectors.toList());

    }

    private Collection<NonSfaExperimentPartController> getNonSfaExperimentPartControllers() {

        Set<NonSfaExperimentPartController> controllers = new HashSet<>();

        for (ExperimentPart experimentPart : experiment.getPartsListCopy()) {
            if (experimentPart instanceof SfaExperimentPart)
                continue;


            ExperimentPartController epc = epcManager.getController(experimentPart);
            if (!(epc instanceof NonSfaExperimentPartController))
                throw new RuntimeException("Expected a non-sfa ExperimentPartController");

            controllers.add((NonSfaExperimentPartController) epc);
        }

        return controllers;
    }

    private boolean hasNonSfaParts() {
        return experiment.getPartsStream().anyMatch(part -> !(part instanceof SfaExperimentPart));
    }

    private boolean hasSfaParts() {
        return experiment.getPartsStream().anyMatch(part -> (part instanceof SfaExperimentPart));
    }
}
